<?php
namespace DEPOClient;

use DEPOClient\Entities\BaseEntity;
use DEPOClient\Entities\Label;
use DEPOClient\Entities\Package;
use DEPOClient\Entities\Place;
use DEPOClient\Exceptions\ConfigException;
use DEPOClient\Exceptions\DEPOException;
use DEPOClient\Exceptions\ConnectionException;

use DEPOClient\Exceptions\InvalidEntityException;
use DEPOClient\Exceptions\NotFoundException;
use GuzzleHttp\Exception\GuzzleException;
use Monolog\Logger;
use Monolog\Handler\StreamHandler;

class Client implements ClientInterface {
	/** @var \GuzzleHttp\Client  */
	protected $client;
	/** @var  Logger */
	protected $log;

	protected $cache_ttl;
	protected $temp_dir;
	protected $pdf_dir;

	protected $log_dir;
	protected $log_level;


	protected $api_base_url;
	protected $api_username;
	protected $api_password;

	public function __construct(array $config){
		$this->_saveConfig($config);
		// create http client
		$this->client = new \GuzzleHttp\Client();
		// create logger
		$this->log = new Logger('DEPOClient');
		$this->log->pushHandler(new StreamHandler($this->log_dir.'/depo.log', $this->log_level));

	}

	protected function _saveConfig(array $config) {
		$configuration = [
			'temp_dir' => __DIR__.'/../temp',
			'log_dir' => __DIR__.'/../log',
			'pdf_dir' => __DIR__.'/../pdf',
			'log_level' => Logger::ERROR,
			'username' => NULL,
			'password' => NULL,
			'base_url' => NULL,
			'cache_ttl' => NULL,
		];
		
		$required = ['username', 'password', 'base_url'];
		
		// update
		foreach($config as $k => $v) {
			$configuration[$k] = $v;
		}
		
		// check required
		foreach($required as $req) {
			if ($configuration[$req] === NULL) {
				throw new ConfigException('Missing configuration: '.$req);
			}
		}
		
		
		// directories
		$this->temp_dir = $configuration['temp_dir'];
		$this->pdf_dir = $configuration['pdf_dir'];
		$this->log_dir = $configuration['log_dir'];

		// create directories if those do not exist
		foreach([$this->temp_dir, $this->pdf_dir, $this->log_level] as $dir) {
			if (!file_exists($dir)) {
				@mkdir($dir, 0777, true);
			}
		}
		
		// api
		$this->api_username = $configuration['username'];
		$this->api_password = $configuration['password'];
		$this->api_base_url = $configuration['base_url'];

		// other
		$this->cache_ttl = intval($configuration['cache_ttl']);
		$this->log_level = $configuration['log_level'];
		
		
	}

	/**
	 * @param $method
	 * @param $uri
	 * @param null $json
	 * @throws DEPOException
	 */
	protected function _jsonRequest($method, $uri, $json = NULL) {
		$options = [
			'headers' => ['Accept' => 'application/json'],
			'json' => $json
		];

		$_body = $this->_generalRequest($method, $uri, $options);
		return json_decode($_body, TRUE);
	}

	/**
	 * Saves pdf as filename.
	 * @param $method
	 * @param $uri
	 * @param $filename
	 * @param null $json
	 * @return string
	 * @throws ConnectionException
	 * @throws DEPOException
	 */
	protected function _pdfRequest($method, $uri, $filename, $json = NULL) {
		$options = [
			'headers' => ['Accept' => 'application/pdf'],
			'json' => $json
		];
		$body = $this->_generalRequest($method, $uri, $options);
		@file_put_contents($filename, $body);
		if (file_exists($filename)) {
			return $filename;
		} else {
			throw new DEPOException('Unable to create pdf file'. $filename);
		}
	}

	/**
	 * Internal wrapper to include exceptions to all requests.
	 * @param $method
	 * @param $uri
	 * @param $options
	 * @return \Psr\Http\Message\StreamInterface
	 * @throws DEPOException
	 */
	protected function _generalRequest($method, $uri, $options) {
		try {
			$url = $this->api_base_url . $uri;
			$options['auth'] = [$this->api_username, $this->api_password];
			$response = $this->client->{strtolower($method)}($url, $options);
			return $response->getBody();
		} catch (GuzzleException $e) {
			// log exception and rethrow it as DEPOException
			$this->log->log(Logger::ERROR, $e->getMessage(), [$e->getFile(), $e->getLine()]);
			if ($e->getCode() === 404) {
				// suppose that 404 generally means bad id
				throw new NotFoundException('Not found: ' . $url);
			} elseif ($e->getCode() >= 400 && $e->getCode() < 500) {
				throw new InvalidEntityException($e->getMessage());
			} else {
				throw new ConnectionException($e->getMessage());
			}
		}
	}

	/**
	 * @param $uri
	 * @param $embeddedKey
	 * @param $conversionClass
	 * @return object[]
	 */
	protected function _pagedGETRequest($uri, $embeddedKey, $conversionClass) {
		$arr = [];
		$page = 1;
		while(TRUE) {
			$resp = $this->_jsonRequest('GET', $uri . ( $page != 1 ? '?page='.$page : '' ) );
			$arr = array_merge($arr, $resp['_embedded'][$embeddedKey]);
			if ($resp['_links']['self']['href'] === $resp['_links']['last']['href']) {
				break;
			}
			$page++;
		}
		$obj = [];
		foreach($arr as $a) {
			/** @var BaseEntity $o */
			$o = new $conversionClass;
			$o->fillFromArray($a);
			$obj[] = $o;
		}
		return $obj;
	}

	/**
	 * @return Place[]
	 * @throws DEPOException
	 */
	public function updatePlacesCache() {
		/** @var Place[] $_places */
		$_places = $this->_pagedGETRequest('/places', 'places', Place::class);
		$places = [];
		foreach($_places as $place) {
			$places[$place->id] = $place;
		}
		// save places as serialized object
		file_put_contents($this->temp_dir . '/places.tmp', serialize($places));
		// save places as json
		file_put_contents($this->temp_dir . '/places.json', json_encode($places, JSON_UNESCAPED_UNICODE));

		return $places;
	}
	
	/**
	 * Reads places from cache / updates cache if empty.
	 * @inheritDoc
	 */
	public function getPlaces(){
		$cachedFilename = $this->temp_dir.'/places.tmp';

		// see if cache exists or is too old
		$cacheExists = @file_exists($cachedFilename);
		$cacheMTime = @filemtime($cachedFilename);
		$cacheOld = ($this->cache_ttl !== NULL) && ( $cacheMTime !== FALSE ) &&
			(time() - $cacheMTime > $this->cache_ttl);
		$refreshCache = !$cacheExists || $cacheOld;

		if ($refreshCache) {
			// update places cache if needed
			$places = $this->updatePlacesCache();
		} else {
			// read places from cache
			$places = @unserialize(file_get_contents($cachedFilename));
			if ($places === FALSE) {
				throw new DEPOException('Unable to read Places from cache.');
			}
		}
		return $places;
	}

	/**
	 * Reads place information directly from API.
	 * @param $id
	 * @return Place
	 * @throws DEPOException
	 */
	protected function _getPlaceViaApi($id) {
		$data = $this->_jsonRequest('GET', '/places/'.$id);
		$place = new Place();
		$place->fillFromArray($data);
		return $place;
	}

	/**
	 * Reads place information from cache.
	 * @param $id
	 * @return Place
	 * @throws DEPOException
	 * @throws NotFoundException
	 */
	protected function _getPlaceFromCache($id) {
		$places = $this->getPlaces();
		if (!isset($places[$id])) {
			// log exception and rethrow it as DEPOException
			$msg = 'Place '.$id.' not found in cache.';
			$this->log->log(Logger::ERROR, $msg);
			throw new NotFoundException($msg);
		} else {
			return $places[$id];
		}
	}

	/**
	 * @inheritdoc
	 */
	public function getPlace($id) {
		return $this->_getPlaceFromCache($id);
	}
	/**
	 * @inheritDoc
	 */
	public function getPackages() {
		return $this->_pagedGETRequest('/packages', 'packages', Package::class);
	}

	/**
	 * @inheritDoc
	 */
	public function getPackage($number){
		$data = $this->_jsonRequest('GET', '/packages/'.$number);
		$package = new Package();
		$package->fillFromArray($data);
		return $package;
	}

	/**
	 * @inheritDoc
	 */
	public function sendPackage(Package $package){
		$package->checkFields();
		$data = $this->_jsonRequest('POST', '/packages/send', $package->asDEPOArray());
		$package->number = $data['number'];
		$package->price = $data['price'];
		return $package;
	}

	/**
	 * @inheritDoc
	 */
	public function cancelPackage($number){
		$this->_jsonRequest('POST', '/packages/cancel', ['number' => $number]);
	}

	/**
	 * @inheritDoc
	 */
	public function generateLabel(Label $label){
		$label->checkFields();
		// pdf filename
		$pdf_filename = $this->pdf_dir.
			'/label_'.$label->number.'_'.$label->size.'.pdf';

		$this->_pdfRequest('POST', '/packages/label', $pdf_filename, $label->asDEPOArray());
		$label->pdf_filename = $pdf_filename;
		return $label;

	}
}